@@ -1,183 +1,91 @@
 """Stateful programmatic WWW navigation, after Perl's WWW::Mechanize.
 
-Copyright 2003-2005 John J. Lee <jjl@pobox.com>
+Copyright 2003-2006 John J. Lee <jjl@pobox.com>
 Copyright 2003 Andy Lester (original Perl code)
 
-This code is free software; you can redistribute it and/or modify it under
-the terms of the BSD License (see the file COPYING included with the
-distribution).
+This code is free software; you can redistribute it and/or modify it
+under the terms of the BSD or ZPL 2.1 licenses (see the file COPYING.txt
+included with the distribution).
 
 """
 
-# XXXX
-# test referer bugs (frags and don't add in redirect unless orig req had Referer)
-
-# XXX
-# The stuff on web page's todo list.
-# Moof's emails about response object, .back(), etc.
-
-from __future__ import generators
-
-import urllib2, urlparse, re, sys
-
-import ClientCookie
-from ClientCookie._Util import response_seek_wrapper
-from ClientCookie._HeadersUtil import split_header_words, is_html
-# serves me right for not using a version tuple...
-VERSION_RE = re.compile(r"(?P<major>\d+)\.(?P<minor>\d+)\.(?P<bugfix>\d+)"
-                        r"(?P<state>[ab])?(?:-pre)?(?P<pre>\d+)?$")
-def parse_version(text):
-    m = VERSION_RE.match(text)
-    if m is None:
-        raise ValueError
-    return tuple([m.groupdict()[part] for part in
-                  ("major", "minor", "bugfix", "state", "pre")])
-assert map(int, parse_version(ClientCookie.VERSION)[:3]) >= [1, 1, 0], \
-       "ClientCookie 1.1.0 or newer is required"
+import urllib2, urlparse, sys, copy, re
 
 from _useragent import UserAgent
+from _html import DefaultFactory
+from _util import response_seek_wrapper, closeable_response
+import _request
 
-__version__ = (0, 0, 11, "a", None)  # 0.0.11a
+__version__ = (0, 1, 2, "b", None)  # 0.1.2b
 
 class BrowserStateError(Exception): pass
 class LinkNotFoundError(Exception): pass
 class FormNotFoundError(Exception): pass
 
-class Link:
-    def __init__(self, base_url, url, text, tag, attrs):
-        assert None not in [url, tag, attrs]
-        self.base_url = base_url
-        self.absolute_url = urlparse.urljoin(base_url, url)
-        self.url, self.text, self.tag, self.attrs = url, text, tag, attrs
-    def __cmp__(self, other):
-        try:
-            for name in "url", "text", "tag", "attrs":
-                if getattr(self, name) != getattr(other, name):
-                    return -1
-        except AttributeError:
-            return -1
-        return 0
-    def __repr__(self):
-        return "Link(base_url=%r, url=%r, text=%r, tag=%r, attrs=%r)" % (
-            self.base_url, self.url, self.text, self.tag, self.attrs)
-
-
-class LinksFactory:
-
-    def __init__(self,
-                 link_parser_class=None,
-                 link_class=Link,
-                 urltags=None,
-                 ):
-        import pullparser
-        assert pullparser.__version__[:3] >= (0, 0, 7), \
-               "pullparser 0.0.7 or newer is required"
-        if link_parser_class is None:
-            link_parser_class = pullparser.TolerantPullParser
-        self.link_parser_class = link_parser_class
-        self.link_class = link_class
-        if urltags is None:
-            urltags = {
-                "a": "href",
-                "area": "href",
-                "frame": "src",
-                "iframe": "src",
-                }
-        self.urltags = urltags
-
-    def links(self, fh, base_url, encoding=None):
-        """Return an iterator that provides links of the document."""
-        import pullparser
-        p = self.link_parser_class(fh, encoding=encoding)
-
-        for token in p.tags(*(self.urltags.keys()+["base"])):
-            if token.data == "base":
-                base_url = dict(token.attrs).get("href")
-                continue
-            if token.type == "endtag":
-                continue
-            attrs = dict(token.attrs)
-            tag = token.data
-            name = attrs.get("name")
-            text = None
-            # XXX need to sort out quoting
-            #url = urllib.quote_plus(attrs.get(self.urltags[tag]))
-            url = attrs.get(self.urltags[tag])
-            if tag == "a":
-                if token.type != "startendtag":
-                    # XXX hmm, this'd break if end tag is missing
-                    text = p.get_compressed_text(("endtag", tag))
-                # but this doesn't work for eg. <a href="blah"><b>Andy</b></a>
-                #text = p.get_compressed_text()
-                # This is a hack from WWW::Mechanize to get some really basic
-                # JavaScript working, which I'm not yet convinced is a good
-                # idea.
-##                 onClick = attrs["onclick"]
-##                 m = re.search(r"/^window\.open\(\s*'([^']+)'/", onClick)
-##                 if onClick and m:
-##                     url = m.group(1)
-            if not url:
-                # Probably an <A NAME="blah"> link or <AREA NOHREF...>.
-                # For our purposes a link is something with a URL, so ignore
-                # this.
-                continue
-
-            yield Link(base_url, url, text, tag, token.attrs)
-
-class FormsFactory:
-
-    """Makes a sequence of objects satisfying ClientForm.HTMLForm interface.
-
-    For constructor argument docs, see ClientForm.ParseResponse
-    argument docs.
 
+class History:
     """
 
-    def __init__(self,
-                 select_default=False,
-                 form_parser_class=None,
-                 request_class=None,
-                 backwards_compat=False,
-                 ):
-        import ClientForm
-        assert map(int, parse_version(ClientForm.VERSION)[:3]) >= [0, 2, 1], \
-               "ClientForm >= 0.2.1a is required"
-        self.select_default = select_default
-        if form_parser_class is None:
-            form_parser_class = ClientForm.FormParser
-        self.form_parser_class = form_parser_class
-        if request_class is None:
-            request_class = ClientCookie.Request
-        self.request_class = request_class
-        self.backwards_compat = backwards_compat
+    Though this will become public, the implied interface is not yet stable.
 
-    def parse_response(self, response):
-        import ClientForm
-        return ClientForm.ParseResponse(
-            response,
-            select_default=self.select_default,
-            form_parser_class=self.form_parser_class,
-            request_class=self.request_class,
-            backwards_compat=self.backwards_compat,
-            )
+    """
+    def __init__(self):
+        self._history = []  # LIFO
+    def add(self, request, response):
+        self._history.append((request, response))
+    def back(self, n, _response):
+        response = _response  # XXX move Browser._response into this class?
+        while n > 0 or response is None:
+            try:
+                request, response = self._history.pop()
+            except IndexError:
+                raise BrowserStateError("already at start of history")
+            n -= 1
+        return request, response
+    def clear(self):
+        del self._history[:]
+    def close(self):
+        for request, response in self._history:
+            response.close()
+        del self._history[:]
 
-    def parse_file(self, file_obj, base_url):
-        import ClientForm
-        return ClientForm.ParseFile(
-            file_obj,
-            base_url,
-            select_default=self.select_default,
-            form_parser_class=self.form_parser_class,
-            request_class=self.request_class,
-            backwards_compat=self.backwards_compat,
-            )
+# Horrible, but needed, at least until fork urllib2.  Even then, may want
+# to preseve urllib2 compatibility.
+def upgrade_response(response):
+    # a urllib2 handler constructed the response, i.e. the response is an
+    # urllib.addinfourl, instead of a _Util.closeable_response as returned
+    # by e.g. mechanize.HTTPHandler
+    try:
+        code = response.code
+    except AttributeError:
+        code = None
+    try:
+        msg = response.msg
+    except AttributeError:
+        msg = None
 
-if sys.version_info[:2] >= (2, 4):
-    from ClientCookie._Opener import OpenerMixin
-else:
-    class OpenerMixin: pass
+    # may have already-.read() data from .seek() cache
+    data = None
+    get_data = getattr(response, "get_data", None)
+    if get_data:
+        data = get_data()
 
-class Browser(UserAgent, OpenerMixin):
+    response = closeable_response(
+        response.fp, response.info(), response.geturl(), code, msg)
+    response = response_seek_wrapper(response)
+    if data:
+        response.set_data(data)
+    return response
+class ResponseUpgradeProcessor(urllib2.BaseHandler):
+    # upgrade responses to be .close()able without becoming unusable
+    handler_order = 0  # before anything else
+    def any_response(self, request, response):
+        if not hasattr(response, 'closeable_response'):
+            response = upgrade_response(response)
+        return response
+
+
+class Browser(UserAgent):
     """Browser-like class with support for history, forms and links.
 
     BrowserStateError is raised whenever the browser is in the wrong state to
@@ -187,56 +95,57 @@ class Browser(UserAgent, OpenerMixin):
 
     Public attributes:
 
-    request: last request (ClientCookie.Request or urllib2.Request)
+    request: current request (mechanize.Request or urllib2.Request)
     form: currently selected form (see .select_form())
-    default_encoding: character encoding used for encoding numeric character
-     references when matching link text, if no encoding is found in the reponse
-     (you should turn on HTTP-EQUIV handling if you want the best chance of
-     getting this right without resorting to this default)
 
     """
 
-    def __init__(self, default_encoding="latin-1",
-                 forms_factory=None,
-                 links_factory=None,
+    handler_classes = UserAgent.handler_classes.copy()
+    handler_classes["_response_upgrade"] = ResponseUpgradeProcessor
+    default_others = copy.copy(UserAgent.default_others)
+    default_others.append("_response_upgrade")
+
+    def __init__(self,
+                 factory=None,
+                 history=None,
                  request_class=None,
                  ):
         """
 
         Only named arguments should be passed to this constructor.
 
-        default_encoding: See class docs.
-        forms_factory: Object supporting the mechanize.FormsFactory interface.
-        links_factory: Object supporting the mechanize.LinksFactory interface.
-        request_class: Request class to use.  Defaults to ClientCookie.Request
+        factory: object implementing the mechanize.Factory interface.
+        history: object implementing the mechanize.History interface.  Note this
+         interface is still experimental and may change in future.
+        request_class: Request class to use.  Defaults to mechanize.Request
          by default for Pythons older than 2.4, urllib2.Request otherwise.
 
-        Note that the supplied forms_factory's request_class attribute is
-        assigned to by this constructor, to ensure only one Request class is
-        used.
+        The Factory and History objects passed in are 'owned' by the Browser,
+        so they should not be shared across Browsers.  In particular,
+        factory.set_response() should not be called except by the owning
+        Browser itself.
+
+        Note that the supplied factory's request_class is overridden by this
+        constructor, to ensure only one Request class is used.
 
         """
-        self.default_encoding = default_encoding
-        self._history = []  # LIFO
+        if history is None:
+            history = History()
+        self._history = history
         self.request = self._response = None
         self.form = None
-        self._forms = None
-        self._title = None
-        self._links = None
 
         if request_class is None:
             if not hasattr(urllib2.Request, "add_unredirected_header"):
-                request_class = ClientCookie.Request
+                request_class = _request.Request
             else:
-                request_class = urllib2.Request  # Python 2.4
+                request_class = urllib2.Request  # Python >= 2.4
+
+        if factory is None:
+            factory = DefaultFactory()
+        factory.set_request_class(request_class)
+        self._factory = factory
         self.request_class = request_class
-        if forms_factory is None:
-            forms_factory = FormsFactory()
-        self._forms_factory = forms_factory
-        forms_factory.request_class = request_class
-        if links_factory is None:
-            links_factory = LinksFactory()
-        self._links_factory = links_factory
 
         UserAgent.__init__(self)  # do this last to avoid __getattr__ problems
 
@@ -244,8 +153,9 @@ class Browser(UserAgent, OpenerMixin):
         if self._response is not None:
             self._response.close()    
         UserAgent.close(self)
-        del self._history[:]
-        self._forms = self._title = self._links = None
+        if self._history is not None:
+            self._history.close()
+            self._history = None
         self.request = self._response = None
 
     def open(self, url, data=None):
@@ -268,24 +178,73 @@ class Browser(UserAgent, OpenerMixin):
                 url = urlparse.urljoin(self._response.geturl(), url)
 
         if self.request is not None and update_history:
-            self._history.append((self.request, self._response))
+            self._history.add(self.request, self._response)
         self._response = None
         # we want self.request to be assigned even if UserAgent.open fails
         self.request = self._request(url, data)
         self._previous_scheme = self.request.get_type()
 
-        self._response = UserAgent.open(self, self.request, data)
-        if not hasattr(self._response, "seek"):
-            self._response = response_seek_wrapper(self._response)
-        self._parse_html(self._response)
+        success = True
+        try:
+            response = UserAgent.open(self, self.request, data)
+        except urllib2.HTTPError, error:
+            success = False
+            response = error
+##         except (IOError, socket.error, OSError), error:
+##             # Yes, urllib2 really does raise all these :-((
+##             # See test_urllib2.py for examples of socket.gaierror and OSError,
+##             # plus note that FTPHandler raises IOError.
+##             # XXX I don't seem to have an example of exactly socket.error being
+##             #  raised, only socket.gaierror...
+##             # I don't want to start fixing these here, though, since this is a
+##             # subclass of OpenerDirector, and it would break old code.  Even in
+##             # Python core, a fix would need some backwards-compat. hack to be
+##             # acceptable.
+##             raise
+        self.set_response(response)
+        if not success:
+            raise error
+        return copy.copy(self._response)
 
-        return self._response
+    def __str__(self):
+        text = []
+        text.append("<%s " % self.__class__.__name__)
+        if self._response:
+            text.append("visiting %s" % self._response.geturl())
+        else:
+            text.append("(not visiting a URL)")
+        if self.form:
+            text.append("\n selected form:\n %s\n" % str(self.form))
+        text.append(">")
+        return "".join(text)
 
     def response(self):
-        """Return last response (as return value of urllib2.urlopen())."""
-        # XXX This is currently broken: responses returned by this method
-        # all share the same seek position.
-        return self._response
+        """Return a copy of the current response.
+
+        The returned object has the same interface as the object returned by
+        .open() (or urllib2.urlopen()).
+
+        """
+        return copy.copy(self._response)
+
+    def set_response(self, response):
+        """Replace current response with (a copy of) response."""
+        # sanity check, necessary but far from sufficient
+        if not (hasattr(response, "info") and hasattr(response, "geturl") and
+                hasattr(response, "read")):
+            raise ValueError("not a response object")
+
+        self.form = None
+
+        if not hasattr(response, "seek"):
+            response = response_seek_wrapper(response)
+        if not hasattr(response, "closeable_response"):
+            response = upgrade_response(response)
+        else:
+            response = copy.copy(response)
+
+        self._response = response
+        self._factory.set_response(self._response)
 
     def geturl(self):
         """Get URL of current document."""
@@ -309,44 +268,22 @@ class Browser(UserAgent, OpenerMixin):
         """
         if self._response is not None:
             self._response.close()
-        while n > 0 or self._response is None:
-            try:
-                self.request, self._response = self._history.pop()
-            except IndexError:
-                raise BrowserStateError("already at start of history")
-            n -= 1
-        self._parse_html(self._response)
-        return self._response
+        self.request, response = self._history.back(n, self._response)
+        self.set_response(response)
+        return response
+
+    def clear_history(self):
+        self._history.clear()
 
     def links(self, **kwds):
         """Return iterable over links (mechanize.Link objects)."""
         if not self.viewing_html():
             raise BrowserStateError("not viewing HTML")
+        links = self._factory.links()
         if kwds:
-            return self._find_links(False, **kwds)
-        if self._links is None:
-            try:
-                self._links = list(self.get_links_iter())
-            finally:
-                self._response.seek(0)
-            return self._links
-
-    def get_links_iter(self):
-        """Return an iterator that provides links of the document.
-
-        This method is provided in addition to .links() to allow lazy iteration
-        over links, while still keeping .links() safe against somebody
-        .seek()ing on a response "behind your back".  When response objects are
-        fixed to have independent seek positions, this method will be
-        deprecated in favour of .links().
-
-        """
-        if not self.viewing_html():
-            raise BrowserStateError("not viewing HTML")
-        base_url = self._response.geturl()
-        self._response.seek(0)
-        return self._links_factory.links(
-            self._response, base_url, self._encoding(self._response))
+            return self._filter_links(links, **kwds)
+        else:
+            return links
 
     def forms(self):
         """Return iterable over forms.
@@ -356,22 +293,19 @@ class Browser(UserAgent, OpenerMixin):
         """
         if not self.viewing_html():
             raise BrowserStateError("not viewing HTML")
-        if self._forms is None:
-            response = self._response
-            response.seek(0)
-            try:
-                self._forms = self._forms_factory.parse_response(response)
-            finally:
-                response.seek(0)
-        return self._forms
+        return self._factory.forms()
 
     def viewing_html(self):
         """Return whether the current response contains HTML data."""
         if self._response is None:
             raise BrowserStateError("not viewing any document")
-        ct_hdrs = self._response.info().getheaders("content-type")
-        url = self._response.geturl()
-        return is_html(ct_hdrs, url)
+        return self._factory.is_html
+
+    def encoding(self):
+        """"""
+        if self._response is None:
+            raise BrowserStateError("not viewing any document")
+        return self._factory.encoding
 
     def title(self):
         """Return title, or None if there is no title element in the document.
@@ -380,27 +314,18 @@ class Browser(UserAgent, OpenerMixin):
         PullParser.get_text() method of pullparser module.
 
         """
-        import pullparser
         if not self.viewing_html():
             raise BrowserStateError("not viewing HTML")
-        if self._title is None:
-            p = pullparser.TolerantPullParser(
-                self._response, encoding=self._encoding(self._response))
-            try:
-                p.get_tag("title")
-            except pullparser.NoMoreTokensError:
-                pass
-            else:
-                self._title = p.get_text()
-        return self._title
+        return self._factory.title
 
     def select_form(self, name=None, predicate=None, nr=None):
         """Select an HTML form for input.
 
         This is a bit like giving a form the "input focus" in a browser.
 
-        If a form is selected, the object supports the HTMLForm interface, so
-        you can call methods like .set_value(), .set(), and .click().
+        If a form is selected, the Browser object supports the HTMLForm
+        interface, so you can call methods like .set_value(), .set(), and
+        .click().
 
         At least one of the name, predicate and nr arguments must be supplied.
         If no matching form is found, mechanize.FormNotFoundError is raised.
@@ -451,7 +376,7 @@ class Browser(UserAgent, OpenerMixin):
         original_scheme = self.request.get_type()
         if scheme not in ["http", "https"]:
             return request
-        if not origin_request and not self.request.has_header('Referer'):
+        if not origin_request and not self.request.has_header("Referer"):
             return request
 
         if (self._handle_referer and
@@ -542,7 +467,8 @@ class Browser(UserAgent, OpenerMixin):
          with opening tags "textified" as per the pullparser docs) must compare
          equal to this argument, if supplied
         text_regex: link text between tag (as defined above) must match the
-         regular expression object passed as this argument, if supplied
+         regular expression object or regular expression string passed as this
+         argument, if supplied
         name, name_regex: as for text and text_regex, but matched against the
          name HTML attribute of the link tag
         url, url_regex: as for text and text_regex, but matched against the
@@ -554,7 +480,10 @@ class Browser(UserAgent, OpenerMixin):
         nr: matches the nth link that matches all other criteria (default 0)
 
         """
-        return self._find_links(True, **kwds)
+        try:
+            return self._filter_links(self._factory.links(), **kwds).next()
+        except StopIteration:
+            raise LinkNotFoundError()
 
     def __getattr__(self, name):
         # pass through ClientForm / DOMForm methods and attributes
@@ -568,7 +497,7 @@ class Browser(UserAgent, OpenerMixin):
 #---------------------------------------------------
 # Private methods.
 
-    def _find_links(self, single,
+    def _filter_links(self, links,
                     text=None, text_regex=None,
                     name=None, name_regex=None,
                     url=None, url_regex=None,
@@ -582,34 +511,22 @@ class Browser(UserAgent, OpenerMixin):
         found_links = []
         orig_nr = nr
 
-        # An optimization, so that if we look for a single link we do not have
-        # to necessarily parse the entire file.
-        if self._links is None and single:
-            all_links = self.get_links_iter()
-        else:
-            if self._links is None:
-                try:
-                    self._links = list(self.get_links_iter())
-                finally:
-                    self._response.seek(0)
-            all_links = self._links
-
-        for link in all_links:
+        for link in links:
             if url is not None and url != link.url:
                 continue
-            if url_regex is not None and not url_regex.search(link.url):
+            if url_regex is not None and not re.search(url_regex, link.url):
                 continue
             if (text is not None and
                 (link.text is None or text != link.text)):
                 continue
             if (text_regex is not None and
-                (link.text is None or not text_regex.search(link.text))):
+                (link.text is None or not re.search(text_regex, link.text))):
                 continue
             if name is not None and name != dict(link.attrs).get("name"):
                 continue
             if name_regex is not None:
                 link_name = dict(link.attrs).get("name")
-                if link_name is None or not name_regex.search(link_name):
+                if link_name is None or not re.search(name_regex, link_name):
                     continue
             if tag is not None and tag != link.tag:
                 continue
@@ -618,28 +535,5 @@ class Browser(UserAgent, OpenerMixin):
             if nr:
                 nr -= 1
                 continue
-            if single:
-                return link
-            else:
-                found_links.append(link)
-                nr = orig_nr
-        if not found_links:
-            raise LinkNotFoundError()
-        return found_links
-
-    def _encoding(self, response):
-        # HTTPEquivProcessor may be in use, so both HTTP and HTTP-EQUIV
-        # headers may be in the response.  HTTP-EQUIV headers come last,
-        # so try in order from first to last.
-        for ct in response.info().getheaders("content-type"):
-            for k, v in split_header_words([ct])[0]:
-                if k == "charset":
-                    return v
-        return self.default_encoding
-
-    def _parse_html(self, response):
-        # this is now lazy, so we just reset the various attributes that
-        # result from parsing
-        self.form = None
-        self._title = None
-        self._forms = self._links = None
+            yield link
+            nr = orig_nr
